/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    classifyImages.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
**              Ian Goodfellow <ia3n@cs.stanford.edu>
** DESCRIPTION:
**  Application for classifying images.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <string.h>

#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#include "svlBase.h"
#include "svlML.h"
#include "svlVision.h"

using namespace std;

#define WINDOW_NAME "classifyImages"

// main -----------------------------------------------------------------------

void usage()
{
	//assert(false);
	cerr << SVL_USAGE_HEADER << endl;
	cerr << "USAGE: ./classifyImages [OPTIONS] <extractor> <model> (<image>* | <imageSequence>)" << endl;
	cerr << "OPTIONS:" << endl
	     << "  -b                :: only search at base scale" << endl
	     << "  -n <object name>  :: name of object" << endl
	     << "  -o <filename>     :: output to file" << endl
	     << "  -r <x> <y> <w> <h> :: scan subregion of image" << endl
	     << "  -R <filename>     :: only scan rectangles in file (<x> <y> <w> <h>)" <<endl
	     << "  -x                :: display image and matches" << endl
	     << SVL_STANDARD_OPTIONS_USAGE 
	     << endl;
}


//[[TODO: classifier type should refer to logistic, boosted, etc.]]

int main(int argc, char *argv[])
{		
  captureOpenCVerrors();
  
  const unsigned MAX_RESULTS_TO_SHOW = 16;
  const unsigned MAX_WIDTH = 1024;

  // read commandline parameters  
  bool bBaseScaleOnly = false;
  const char *objectName = "[unknown]";
  const char *outputFilename = NULL;
  CvRect subRegion = cvRect(0, 0, 0, 0);
  char *regionsFile = NULL;
  bool bDisplayImage = false;
  
  SVL_BEGIN_CMDLINE_PROCESSING(argc, argv)
    SVL_CMDLINE_BOOL_OPTION("-b", bBaseScaleOnly)
    SVL_CMDLINE_STR_OPTION("-n", objectName)
    SVL_CMDLINE_STR_OPTION("-o", outputFilename)
    SVL_CMDLINE_OPTION_BEGIN("-r", ptr)
      subRegion.x = atoi(ptr[0]);
      subRegion.y = atoi(ptr[1]);
      subRegion.width = atoi(ptr[2]);
      subRegion.height = atoi(ptr[3]);
    SVL_CMDLINE_OPTION_END(4)
    SVL_CMDLINE_STR_OPTION("-R", regionsFile)
    SVL_CMDLINE_BOOL_OPTION("-x", bDisplayImage)

   // [[TODO: deprecate from here]] --------------------------------------
    else if (!strcmp(*SVL_CMDLINE_ARGV, "-ch")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlImageLoader channels")
      string type = string(*(++SVL_CMDLINE_ARGV));
      string ext = string(*(++SVL_CMDLINE_ARGV));
      pair<string, string> ch = make_pair<string, string>(type, ext);
      svlImageLoader::CHANNELS.push_back(ch);
      SVL_CMDLINE_ARGC -= 2;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-dmsize")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlImageLoader depthMapWidth and depthMapHeight")
      svlImageLoader::DEPTH_MAP_WIDTH = atoi(*(++SVL_CMDLINE_ARGV));
      svlImageLoader::DEPTH_MAP_HEIGHT = atoi(*(++SVL_CMDLINE_ARGV));
      SVL_CMDLINE_ARGC -= 2;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-s")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlImageLoader resizeWidth and resizeHeight")
      svlImageLoader::WIDTH = atoi(*(++SVL_CMDLINE_ARGV));
      svlImageLoader::HEIGHT = atoi(*(++SVL_CMDLINE_ARGV));
      SVL_CMDLINE_ARGC -= 2;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-baseExt")) {
      SVL_OPTION_DEPRECATED(*SVL_CMDLINE_ARGV, "svlVision.svlImageLoader defaultExtension")
      svlImageLoader::BASE_EXT = *(++SVL_CMDLINE_ARGV);
      SVL_CMDLINE_ARGV--;
    } else if (!strcmp(*SVL_CMDLINE_ARGV, "-c")) {
      SVL_LOG(SVL_LOG_MESSAGE, "-c has deprecated and is being ignored entirely. "
	      << "The default behavior has not changed. In the future, -c will refer to "
	      << "the classifier type, e.g. logistic, boosted, etc. You should delete this "
	      << "option from your scripts as to not cause confusion later");
    }


    SVL_CMDLINE_OPTION_BEGIN("-t", ptr)
      SVL_OPTION_DEPRECATED("-t", "svlVision.svlSlidingWindowDetector threshold")
      svlSlidingWindowDetector::THRESHOLD = atof(ptr[0]);
    SVL_CMDLINE_OPTION_END(1)
    SVL_CMDLINE_OPTION_BEGIN("-delta_scale", ptr)
      SVL_OPTION_DEPRECATED("-delta_scale", "svlVision.svlSlidingWindowDetector deltaScale")
      svlSlidingWindowDetector::DELTA_SCALE = atof(ptr[0]);
    SVL_CMDLINE_OPTION_END(1)
    SVL_CMDLINE_OPTION_BEGIN("-delta_shift", ptr)
      SVL_OPTION_DEPRECATED("-delta_shift", "svlVision.svlSlidingWindowDetector deltaX and deltaY")
      svlSlidingWindowDetector::DELTA_X = svlSlidingWindowDetector::DELTA_Y = atoi(ptr[0]);
    SVL_CMDLINE_OPTION_END(1)

    // [[TODO: deprecate to here]] --------------------------------------

  SVL_END_CMDLINE_PROCESSING(usage());
    
  if (SVL_CMDLINE_ARGC < 3) {
    usage();
    return -1;
  }

  svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));
	
  const char *extractorFilename = *SVL_CMDLINE_ARGV++; SVL_CMDLINE_ARGC--;
  const char *classifierFilename = *SVL_CMDLINE_ARGV++; SVL_CMDLINE_ARGC--;

  int classifyHandle = svlCodeProfiler::getHandle("classification");

  svlImageLoader loader;
  loader.readCommandLine(SVL_CMDLINE_ARGV, SVL_CMDLINE_ARGC);

  svlSlidingWindowDetector detector(objectName);
  assert(detector.load(extractorFilename, classifierFilename));
  
  // run classifier on all supplied images
  unsigned numImages = loader.numFrames();
  vector<IplImage *> annotatedImages;
  annotatedImages.resize(((numImages < MAX_RESULTS_TO_SHOW) ?
			  numImages : MAX_RESULTS_TO_SHOW), NULL);
  if (sqrt((double)annotatedImages.size()) * loader.getResizeWidth() > MAX_WIDTH) {
    annotatedImages.resize(1);
  }
  IplImage *resultsImage = NULL;
  
  if (bDisplayImage) {
    cvNamedWindow(WINDOW_NAME, 0);
  }

  ofstream ofs;
  if (outputFilename) {
    ofs.open(outputFilename);
    assert(!ofs.fail());
    ofs << "<Object2dSequence>" << endl;
  }
 
  SVL_LOG(SVL_LOG_MESSAGE, "Running classifier on " << numImages << " images/sequences...");

  double threshold = detector.getThreshold();
 
  for (unsigned index = 0; index < numImages; index++) 
    {
      SVL_LOG(SVL_LOG_VERBOSE, "Processing " << loader.getFrameBasename(index) << "...");

      vector<IplImage *> images;
      bool success = loader.getFrame(index, images);
      if (!success) continue;

      // classify image
      svlObject2dFrame objects;

      svlCodeProfiler::tic(classifyHandle);
      if (regionsFile != NULL) {
          CvRect r;
          ifstream ifs(regionsFile);
          SVL_ASSERT(!ifs.fail());
          while (1) {
              ifs >> r.x >> r.y >> r.width >> r.height;
              if (ifs.fail()) break;
              svlObject2d o(r);
              o.pr = detector.evaluateModel(images, r);
              o.pr = 1.0 / (1.0 + exp(-o.pr));
              o.name = string(objectName);
              objects.push_back(o);
          }
          ifs.close();
      } else if (bBaseScaleOnly) {
	  vector<CvPoint> locations;
	  if ((subRegion.width > 0) && (subRegion.height > 0)) {
	    detector.createWindowLocations(subRegion.width, subRegion.height, locations);
	    for (vector<CvPoint>::iterator it = locations.begin(); it != locations.end(); ++it) {
	      it->x += subRegion.x;
	      it->y += subRegion.y;
	    }
	  } 
	  else 
	    {
	      detector.createWindowLocations(images[0]->width, images[0]->height, locations);
	    }
	  	detector.classifyImage(images, locations, &objects);
	} 
      else if ((subRegion.width > 0) && (subRegion.height > 0)) {
				detector.classifySubImage(images, subRegion, &objects);
      } 
      else {
				detector.classifyImage(images, &objects);
      }
      svlCodeProfiler::toc(classifyHandle);
     
      // output objects
      if (outputFilename) {
	writeObject2dFrame(ofs, objects, strBaseName(loader.getFrameBasename(index)).c_str());
      }
     
      if (bDisplayImage) {
	IplImage *colourImage = cvCreateImage(cvSize(images[0]->width, images[0]->height),
					      IPL_DEPTH_8U, 3);
	if (images[0]->nChannels == 1) {
	  cvCvtColor(images[0], colourImage, CV_GRAY2RGB);
	} else if (images[0]->nChannels == 3) {
	  cvCopy(images[0], colourImage);
	}
	if (resultsImage != NULL) {	
	  cvReleaseImage(&resultsImage);
	}
       
	for (unsigned i = 0; i < objects.size(); i++) {
	  cvRectangle(colourImage, cvPoint((int)objects[i].x, (int)objects[i].y),
		      cvPoint((int)(objects[i].x + objects[i].w - 1), 
			      (int)(objects[i].y + objects[i].h - 1)),
		      CV_RGB(0, (unsigned char)(255 * (objects[i].pr - threshold) / (1.0 - threshold)), 0), 3, 8);
	  cvRectangle(colourImage, cvPoint((int)objects[i].x, (int)objects[i].y),
		      cvPoint((int)(objects[i].x + objects[i].w - 1), 
			      (int)(objects[i].y + objects[i].h - 1)),
		      CV_RGB(255, 255, 255), 1);
	}
       
	cvRectangle(colourImage, cvPoint(subRegion.x, subRegion.y),
		    cvPoint(subRegion.x + subRegion.width - 1, subRegion.y + subRegion.height - 1), 
		    CV_RGB(0, 0, 255), 1);
       
	if (annotatedImages[index % annotatedImages.size()] != NULL) {
	  cvReleaseImage(&annotatedImages[index % annotatedImages.size()]);
	}
	annotatedImages[index % annotatedImages.size()] = colourImage;
	resultsImage = combineImages(annotatedImages);
	cvShowImage(WINDOW_NAME, resultsImage);
	int c = cvWaitKey(index != numImages - 1 ? 100 : 0);
	//int c = cvWaitKey(0);
       
	if (c == (int)'s') {
	  stringstream filename;
	  filename << "classifyImage" << index << ".jpg";
	  cvSaveImage(filename.str().c_str(), resultsImage);
	}
      }
     
      // free image
      for (unsigned i = 0; i < images.size(); i++) {
	cvReleaseImage(&images[i]);
      }
     
    }//image names in image sequence loop
 
  if (outputFilename) {
    ofs << "</Object2dSequence>" << endl;
    ofs.close();
  }
 
  // free results
  if (resultsImage != NULL) {
    cvReleaseImage(&resultsImage);
  }
  for (unsigned i = 0; i < annotatedImages.size(); i++) {
    if (annotatedImages[i] != NULL) {
      cvReleaseImage(&annotatedImages[i]);
    }
  }
 

  svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
  svlCodeProfiler::print(cerr);
 
  return 0;
}



