/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlHistogramMemManager.h
** AUTHOR(S):   Olga Russakovsky <olga@cs.stanford.edu>
** DESCRIPTION:
**   Defines a general feature vector represented as a vector of doubles.
** Useful for e.g. svlHistogram. Be careful to account for _offset when
** implementing subclasses.
**
*****************************************************************************/

#pragma once

#include <string>
#include <vector>
#include <fstream>
#include <cv.h>
#include "svlBase.h"

using namespace std;

// svlHistogramMemManager ---------------------------------------------------------
class svlHistogramMemManager
{
 protected:
  // optionally provided by user or constructed
  vector<double> *_values;
  unsigned _offset; // will only concern itself with values between _offset
               // and _offset + _numValues

 private:
  unsigned _numValues; // used to resize vectors supplied later
  bool _ownsValues;  // if true, will deallocate _values at the end

 public:
  // Creates a feature vector, either allocating it internally or
  // expecting it to be supplied later via the 'set' method
  svlHistogramMemManager(unsigned num, bool allocateSpace = false);
  svlHistogramMemManager(svlHistogramMemManager &m);
  ~svlHistogramMemManager();
  
  // Will resize and clear the feature vector
  void set(vector<double> *fv, unsigned offset = 0);

  void clear();
  void normalize(); // todo: not only unit length
 // Will truncate all entires > t to t
  void truncate(double t);

  // Will return the i^th element (accounting for offset!)
  inline double operator[](unsigned i) const {
    SVL_ASSERT_MSG(_values, "storage for the histogram must be provided");
    SVL_ASSERT(i >= 0 && i < (unsigned)_numValues); // todo: more user friendly
  
    return _values->at(i+_offset);
  }

 protected:
  inline double &elem(unsigned i) {
    SVL_ASSERT_MSG(_values, "storage for the histogram must be provided");
    SVL_ASSERT(i >= 0 && i < (unsigned)_numValues); // todo: more user friendly
  
    return _values->at(i+_offset);
  }
};
